#
# Copyright (C) 2018, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl
#
# Please see LICENSE.txt file included in the top-level directory of the
# appropriate download for a copy of the license and additional information.
#

from base64 import b64encode
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import SHA256
from datetime import datetime
from locale import LC_ALL, setlocale
from os import path
from requests import Session
from threading import Timer
try:
    import oci
except ImportError:
    oci = None

from borneo.auth import AuthorizationProvider
from borneo.common import CheckValue, HttpConstants, LogUtils, Memoize
from borneo.exception import IllegalArgumentException
from borneo.http import RequestUtils


class SignatureProvider(AuthorizationProvider):
    """
    Cloud service only.

    An instance of :py:class:`borneo.AuthorizationProvider` that generates and
    caches signature for each request as authorization string. A number of
    pieces of information are required for configuration. See `Required Keys and
    OCIDs <https://docs.cloud.oracle.com/iaas/Content/API/Concepts/
    apisigningkey.htm>`_ for information and instructions on how to create the
    required keys and OCIDs for configuration. The required information
    includes:

        * A signing key, used to sign requests.
        * A pass phrase for the key, if it is encrypted.
        * The fingerprint of the key pair used for signing.
        * The OCID of the tenancy.
        * The OCID of a user in the tenancy.

    All of this information is required to authenticate and authorize access to
    the service. See :ref:`creds-label` for information on how to acquire this
    information.

    There are two different ways to authorize an application:

    1. Using a specific user's identity.
    2. Using an Instance Principal, which can be done when running on a compute
       instance in the Oracle Cloud Infrastructure (OCI). See
       :py:meth:`create_with_instance_principal` and `Calling Services from
       Instances <https://docs.cloud.oracle.com/iaas/Content/Identity/Tasks/
       callingservicesfrominstances.htm>`_.

    The latter can be simpler to use when running on an OCI compute instance,
    but limits the ability to use a compartment name vs OCID when naming
    compartments and tables in :py:class:`Request` classes and when naming
    tables in queries. A specific user identity is best for naming flexibility,
    allowing both compartment names and OCIDs.

    When using a specific user's identity there are 3 options for providing the
    required information:

    1. Using a instance of oci.signer.Signer or
       oci.auth.signers.InstancePrincipalsSecurityTokenSigner
    2. Directly providing the credentials via parameters
    3. Using a configuration file

    Only one method of providing credentials can be used, and if they are mixed
    the priority from high to low is:

    * Signer or InstancePrincipalsSecurityTokenSigner(provider is used)
    * Credentials as arguments (tenant_id, etc used)
    * Configuration file (config_file is used)

    :param provider: an instance of oci.signer.Signer or
        oci.auth.signers.InstancePrincipalsSecurityTokenSigner.
    :type provider: Signer or InstancePrincipalsSecurityTokenSigner
    :param config_file: path of configuration file.
    :type config_file: str
    :param profile_name: user profile name. Only valid with config_file.
    :type profile_name: str
    :param tenant_id: id of the tenancy
    :type tenant_id: str
    :param user_id: id of a specific user
    :type user_id: str
    :param private_key: path to private key or private key content
    :type private_key: str
    :param fingerprint: fingerprint for the private key
    :type fingerprint: str
    :param pass_phrase: pass_phrase for the private key if created
    :type pass_phrase: str
    :param duration_seconds: the signature cache duration in seconds.
    :type duration_seconds: int
    :param refresh_ahead: the refresh time before signature cache expiry
       in seconds.
    :type refresh_ahead: int
    :raises IllegalArgumentException: raises the exception if the parameters
        are not valid.
    """

    SIGNING_HEADERS = '(request-target) host date'
    CACHE_KEY = 'signature'
    """Cache key name."""
    MAX_ENTRY_LIFE_TIME = 300
    """Maximum lifetime of signature 300 seconds."""
    DEFAULT_REFRESH_AHEAD = 10
    """Default refresh time before signature expiry, 10 seconds."""
    SIGNATURE_HEADER_FORMAT = (
        'Signature headers="{0}",keyId="{1}",algorithm="{2}",signature="{3}",' +
        'version="{4}"')
    SIGNATURE_VERSION = 1

    def __init__(self, provider=None, config_file=None, profile_name=None,
                 tenant_id=None, user_id=None, fingerprint=None,
                 private_key=None, pass_phrase=None,
                 duration_seconds=MAX_ENTRY_LIFE_TIME,
                 refresh_ahead=DEFAULT_REFRESH_AHEAD):
        """
        The SignatureProvider that generates and caches request signature.
        """
        CheckValue.check_int_gt_zero(duration_seconds, 'duration_seconds')
        CheckValue.check_int_gt_zero(refresh_ahead, 'refresh_ahead')
        if duration_seconds > SignatureProvider.MAX_ENTRY_LIFE_TIME:
            raise IllegalArgumentException(
                'Access token cannot be cached longer than ' +
                str(SignatureProvider.MAX_ENTRY_LIFE_TIME) + ' seconds.')

        #
        # This class depends on the oci package
        #
        if oci is None:
            raise ImportError('Package \'oci\' is required; please install')

        try:
            if provider is not None:
                if not isinstance(
                    provider,
                    (oci.signer.Signer,
                     oci.auth.signers.InstancePrincipalsSecurityTokenSigner)):
                    raise IllegalArgumentException(
                        'provider should be an instance of oci.signer.Signer' +
                        'or oci.auth.signers.' +
                        'InstancePrincipalsSecurityTokenSigner.')
                self._provider = provider
            elif (tenant_id is None or user_id is None or fingerprint is None or
                  private_key is None):
                CheckValue.check_str(config_file, 'config_file', True)
                CheckValue.check_str(profile_name, 'profile_name', True)
                if config_file is None and profile_name is None:
                    # Use default user profile and private key from default path
                    # of configuration file ~/.oci/config.
                    config = oci.config.from_file()
                elif config_file is None and profile_name is not None:
                    # Use user profile with given profile name and private key
                    # from default path of configuration file ~/.oci/config.
                    config = oci.config.from_file(profile_name=profile_name)
                elif config_file is not None and profile_name is None:
                    # Use user profile with default profile name and private key
                    # from specified configuration file.
                    config = oci.config.from_file(file_location=config_file)
                else:  # config_file is not None and profile_name is not None
                    # Use user profile with given profile name and private key
                    # from specified configuration file.
                    config = oci.config.from_file(
                        file_location=config_file, profile_name=profile_name)
                self._provider = oci.signer.Signer(
                    config['tenancy'], config['user'], config['fingerprint'],
                    config['key_file'], config.get('pass_phrase'),
                    config.get('key_content'))
            else:
                CheckValue.check_str(tenant_id, 'tenant_id')
                CheckValue.check_str(user_id, 'user_id')
                CheckValue.check_str(fingerprint, 'fingerprint')
                CheckValue.check_str(private_key, 'private_key')
                CheckValue.check_str(pass_phrase, 'pass_phrase', True)
                if path.isfile(private_key):
                    key_file = private_key
                    key_content = None
                else:
                    key_file = None
                    key_content = private_key
                self._provider = oci.signer.Signer(
                    tenant_id, user_id, fingerprint, key_file, pass_phrase,
                    key_content)
        except AttributeError:
            raise ImportError('Package \'oci\' is required; please install')
        self._signature_cache = Memoize(duration_seconds)
        self._refresh_interval_s = (duration_seconds - refresh_ahead if
                                    duration_seconds > refresh_ahead else 0)

        # Refresh timer.
        self._timer = None
        self._service_host = None
        self._logger = None
        self._logutils = LogUtils()
        self._sess = Session()
        self._request_utils = RequestUtils(self._sess, self._logutils)

    def close(self):
        """
        Closes the signature provider.
        """
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None

    def get_authorization_string(self, request=None):
        if self._service_host is None:
            raise IllegalArgumentException(
                'Unable to find service host, use set_service_host to load ' +
                'from NoSQLHandleConfig')
        sig_details = self._get_signature_details()
        if sig_details is not None:
            return sig_details.get_signature_header()

    def get_logger(self):
        return self._logger

    def set_logger(self, logger):
        CheckValue.check_logger(logger, 'logger')
        self._logger = logger
        self._logutils = LogUtils(logger)
        self._request_utils = RequestUtils(self._sess, self._logutils)
        return self

    def set_required_headers(self, request, auth_string, headers):
        sig_details = self._get_signature_details()
        if sig_details is None:
            return
        headers[HttpConstants.AUTHORIZATION] = (
            sig_details.get_signature_header())
        headers[HttpConstants.DATE] = sig_details.get_date()
        compartment = request.get_compartment()
        if compartment is None:
            # If request doesn't has compartment, set the tenant id as the
            # default compartment, which is the root compartment in IAM if using
            # user principal. If using an instance principal this value is
            # None.
            compartment = self._get_tenant_ocid()
        if compartment is not None:
            headers[HttpConstants.REQUEST_COMPARTMENT_ID] = compartment
        else:
            raise IllegalArgumentException(
                'Compartment is None. When authenticating using an Instance ' +
                'Principal the compartment for the operation must be specified.'
            )

    def set_service_host(self, config):
        service_url = config.get_service_url()
        if service_url is None:
            raise IllegalArgumentException('Must set service URL first.')
        self._service_host = service_url.hostname
        return self

    @staticmethod
    def create_with_instance_principal(iam_auth_uri=None):
        """
        Creates a SignatureProvider using an instance principal. This method may
        be used when calling the Oracle NoSQL Database Cloud Service from an
        Oracle Cloud compute instance. It authenticates with the instance
        principal and uses a security token issued by IAM to do the actual
        request signing.

        When using an instance principal the compartment (OCID) must be
        specified on each request or defaulted by using
        :py:meth:`borneo.NoSQLHandleConfig.set_default_compartment`. If the
        compartment is not specified for an operation an exception will be
        thrown.

        See `Calling Services from Instances <https://docs.cloud.oracle.com/
        iaas/Content/Identity/Tasks/callingservicesfrominstances.htm>`_

        :param iam_auth_uri: the URI is usually detected automatically, specify
            the URI if you need to overwrite the default, or encounter the
            *Invalid IAM URI* error, it is optional.
        :type iam_auth_uri: str
        :returns: a SignatureProvider.
        :rtype: SignatureProvider
        """
        if iam_auth_uri is None:
            return SignatureProvider(
                oci.auth.signers.InstancePrincipalsSecurityTokenSigner())
        else:
            return SignatureProvider(
                oci.auth.signers.InstancePrincipalsSecurityTokenSigner(
                    federation_endpoint=iam_auth_uri))

    def _get_signature_details(self):
        sig_details = self._signature_cache.get(SignatureProvider.CACHE_KEY)
        if sig_details is not None:
            return sig_details
        sig_details = self._get_signature_details_internal()
        self._signature_cache.set(SignatureProvider.CACHE_KEY, sig_details)
        self._schedule_refresh()
        return sig_details

    def _get_signature_details_internal(self):
        setlocale(LC_ALL, 'en_US')
        date_str = datetime.utcnow().strftime('%a, %d %b %Y %H:%M:%S GMT')
        if isinstance(self._provider,
                      oci.auth.signers.InstancePrincipalsSecurityTokenSigner):
            self._provider.refresh_security_token()
        private_key = self._provider.private_key
        key_id = self._provider.api_key

        try:
            signature = private_key.sign(
                self._signing_content(date_str), PKCS1v15(), SHA256())
            signature = b64encode(signature)
        except TypeError:
            signature = private_key.sign(
                self._signing_content(date_str).encode(), PKCS1v15(), SHA256())
            signature = b64encode(signature).decode()
        sig_header = str.format(
            SignatureProvider.SIGNATURE_HEADER_FORMAT,
            SignatureProvider.SIGNING_HEADERS, key_id, 'rsa-sha256', signature,
            SignatureProvider.SIGNATURE_VERSION)
        return SignatureProvider.SignatureDetails(sig_header, date_str)

    def _get_tenant_ocid(self):
        """
        Get tenant OCID if using user principal.

        :returns: tenant OCID of user.
        :rtype: str
        """
        if isinstance(self._provider, oci.signer.Signer):
            return self._provider.api_key.split('/')[0]

    def _refresh_task(self):
        try:
            sig_details = self._get_signature_details_internal()
            if sig_details is not None:
                self._signature_cache.set(SignatureProvider.CACHE_KEY,
                                          sig_details)
                self._schedule_refresh()
        except Exception as e:
            # Ignore the failure of refresh. The driver would try to generate
            # signature in the next request if signature is not available, the
            # failure would be reported at that moment.
            self._logutils.log_warning(
                'Unable to refresh cached request signature, ' + str(e))
            self._timer.cancel()
            self._timer = None

    def _schedule_refresh(self):
        # If refresh interval is 0, don't schedule a refresh.
        if self._refresh_interval_s == 0:
            return
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None
        self._timer = Timer(self._refresh_interval_s, self._refresh_task)
        self._timer.start()

    def _signing_content(self, date_str):
        return (HttpConstants.REQUEST_TARGET + ': post /' +
                HttpConstants.NOSQL_DATA_PATH + '\nhost: ' +
                self._service_host + '\ndate: ' + date_str)

    class SignatureDetails(object):
        def __init__(self, signature_header, date_str):
            # Signing date, keep it and pass along with each request, so
            # requests can reuse the signature within the 5-mins time window.
            self._date = date_str
            # Signature header string.
            self._signature_header = signature_header

        def get_date(self):
            return self._date

        def get_signature_header(self):
            return self._signature_header

        def is_valid(self, header):
            if header is None:
                return False
            return header == self._signature_header



#
# Copyright (C) 2018, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl
#
# Please see LICENSE.txt file included in the top-level directory of the
# appropriate download for a copy of the license and additional information.
#

from logging import DEBUG
from requests import ConnectionError, Timeout, codes
from threading import Lock
from time import time

from .common import ByteInputStream, HttpConstants, synchronized
from .exception import (
    IllegalStateException, NoSQLException, RequestTimeoutException,
    RetryableException, SecurityInfoNotReadyException)
from .kv import AuthenticationException, StoreAccessTokenProvider
from .serde import BinaryProtocol


class HttpResponse(object):
    # Class to package HTTP response output and status code.
    def __init__(self, content, status_code):
        self._content = content
        self._status_code = status_code

    def __str__(self):
        return ('HttpResponse [content=' + self._content + ', status_code=' +
                str(self._status_code) + ']')

    def get_content(self):
        return self._content

    def get_status_code(self):
        return self._status_code


class RequestUtils(object):
    # Utility to issue http request.
    def __init__(self, sess, logutils, request=None, retry_handler=None,
                 client=None):
        """
        Init the RequestUtils.

        :param sess: the session.
        :type sess: Session
        :param logutils: contains the logging methods.
        :type logutils: LogUtils
        :param request: request to execute.
        :type request: Request
        :param retry_handler: the retry handler.
        :type retry_handler: RetryHandler
        """
        self._sess = sess
        self._logutils = logutils
        self._request = request
        self._retry_handler = retry_handler
        self._client = client
        self._max_request_id = 1
        self._auth_provider = (
            client.get_auth_provider() if client is not None else None)
        self.lock = Lock()

    def do_delete_request(self, uri, headers, timeout_ms):
        """
        Issue HTTP DELETE request with retries and general error handling.

        It retries upon seeing following exceptions and response codes:

            IOException\n
            HTTP response with status code larger than 500\n
            Other throwable excluding RuntimeException, InterruptedException,
            ExecutionException and TimeoutException.

        :param uri: the request URI.
        :type uri: str
        :param headers: HTTP headers of this request.
        :type headers: dict
        :param timeout_ms: request timeout in milliseconds.
        :type timeout_ms: int
        :returns: HTTP response, a object encapsulate status code and response.
        :rtype: HttpResponse or Result
        """
        return self._do_request('DELETE', uri, headers, None, timeout_ms,
                                sec_timeout_ms=0)

    def do_get_request(self, uri, headers, timeout_ms):
        """
        Issue HTTP GET request with retries and general error handling.

        It retries upon seeing following exceptions and response codes:

            IOException\n
            HTTP response with status code larger than 500\n
            Other throwable excluding RuntimeException, InterruptedException,
            ExecutionException and TimeoutException.

        :param uri: the request URI.
        :type uri: str
        :param headers: HTTP headers of this request.
        :type headers: dict
        :param timeout_ms: request timeout in milliseconds.
        :type timeout_ms: int
        :returns: HTTP response, a object encapsulate status code and response.
        :rtype: HttpResponse or Result
        """
        return self._do_request('GET', uri, headers, None, timeout_ms,
                                sec_timeout_ms=0)

    def do_post_request(self, uri, headers, payload, timeout_ms,
                        sec_timeout_ms=0):
        """
        Issue HTTP POST request with retries and general error handling.

        It retries upon seeing following exceptions and response codes:

            IOException\n
            HTTP response with status code larger than 500\n
            Other throwable excluding RuntimeException, InterruptedException,
            ExecutionException and TimeoutException

        :param uri: the request URI.
        :type uri: str
        :param headers: HTTP headers of this request.
        :type headers: dict
        :param payload: payload in string.
        :type payload: str
        :param timeout_ms: request timeout in milliseconds.
        :type timeout_ms: int
        :param sec_timeout_ms: the timeout of waiting security information to be
            available in milliseconds.
        :type sec_timeout_ms: int
        :returns: HTTP response, a object encapsulate status code and response.
        :rtype: HttpResponse or Result
        """
        return self._do_request('POST', uri, headers, payload, timeout_ms,
                                sec_timeout_ms)

    def do_put_request(self, uri, headers, payload, timeout_ms,
                       sec_timeout_ms=0):
        """
        Issue HTTP PUT request with retries and general error handling.

        It retries upon seeing following exceptions and response codes:

            IOException\n
            HTTP response with status code larger than 500\n
            Other throwable excluding RuntimeException, InterruptedException,
            ExecutionException and TimeoutException

        :param uri: the request URI.
        :type uri: str
        :param headers: HTTP headers of this request.
        :type headers: dict
        :param payload: payload in string.
        :type payload: str
        :param timeout_ms: request timeout in milliseconds.
        :type timeout_ms: int
        :param sec_timeout_ms: the timeout of waiting security information to be
            available in milliseconds.
        :type sec_timeout_ms: int
        :returns: HTTP response, a object encapsulate status code and response.
        :rtype: HttpResponse or Result
        """
        return self._do_request('PUT', uri, headers, payload, timeout_ms,
                                sec_timeout_ms)

    def _do_request(self, method, uri, headers, payload, timeout_ms,
                    sec_timeout_ms):
        start_ms = int(round(time() * 1000))
        timeout_s = float(timeout_ms) / 1000
        throttle_retried = 0
        num_retried = 0
        exception = None
        while True:
            if self._auth_provider is not None:
                auth_string = self._auth_provider.get_authorization_string(
                    self._request)
                self._auth_provider.validate_auth_string(auth_string)
                self._auth_provider.set_required_headers(
                    self._request, auth_string, headers)
            if num_retried > 0:
                self._log_retried(num_retried, exception)
            response = None
            try:
                if self._request is not None:
                    request_id = str(self._next_request_id())
                    headers[HttpConstants.REQUEST_ID_HEADER] = request_id
                elif payload is not None:
                    payload = payload.encode()
                if payload is None:
                    response = self._sess.request(
                        method, uri, headers=headers, timeout=timeout_s)
                else:
                    # wrap the payload to make it compatible with pyOpenSSL,
                    # there maybe small cost to wrap it.
                    response = self._sess.request(
                        method, uri, headers=headers, data=memoryview(payload),
                        timeout=timeout_s)
                if self._logutils.is_enabled_for(DEBUG):
                    self._logutils.log_debug(
                        'Response: ' + self._request.__class__.__name__ +
                        ', status: ' + str(response.status_code))
                if self._request is not None:
                    return self._process_response(
                        self._request, response.content, response.status_code)
                else:
                    res = HttpResponse(response.content.decode(),
                                       response.status_code)
                    """
                    Retry upon status code larger than 500, in general, this
                    indicates server internal error.
                    """
                    if res.get_status_code() >= codes.server_error:
                        self._logutils.log_debug(
                            'Remote server temporarily unavailable, status ' +
                            'code ' + str(res.get_status_code()) +
                            ' , response ' + res.get_content())
                        num_retried += 1
                        continue
                    return res
            except AuthenticationException as ae:
                if (self._auth_provider is not None and isinstance(
                        self._auth_provider, StoreAccessTokenProvider)):
                    self._auth_provider.bootstrap_login()
                    exception = ae
                    num_retried += 1
                    continue
                self._logutils.log_info(
                    'Unexpected authentication exception: ' + str(ae))
                raise NoSQLException('Unexpected exception: ' + str(ae), ae)
            except RetryableException as re:
                self._logutils.log_debug('Retryable exception: ' + str(re))
                """
                Handle automatic retries. If this does not throw an error, then
                the delay (if any) will have been performed and the request
                should be retried.

                If there have been too many retries this method will throw the
                original exception.
                """
                retried = self._handle_retry(
                    re, self._request, throttle_retried)
                """
                Don't count retries for security info not ready as throttle
                retires.
                """
                if not isinstance(re, SecurityInfoNotReadyException):
                    throttle_retried = retried
                exception = re
                num_retried += 1
            except NoSQLException as nse:
                self._logutils.log_error(
                    'Client execution NoSQLException: ' + str(nse))
                raise nse
            except RuntimeError as re:
                self._logutils.log_error(
                    'Client execution RuntimeError: ' + str(re))
                raise re
            except ConnectionError as ce:
                self._logutils.log_error(
                    'HTTP request execution ConnectionError: ' + str(ce))
                raise ce
            except Timeout as t:
                if self._request is not None:
                    self._logutils.log_error('Timeout exception: ' + str(t))
                    break  # fall through to exception below
                raise RuntimeError('Timeout exception: ' + str(t))
            finally:
                if response is not None:
                    response.close()
            if self._timeout_request(
                    start_ms, timeout_ms, sec_timeout_ms, exception):
                break
        actual_timeout = timeout_ms
        if (self._request is not None and
                isinstance(exception, SecurityInfoNotReadyException)):
            actual_timeout = sec_timeout_ms
        raise RequestTimeoutException(
            'Request timed out after ' + str(num_retried) +
            (' retry.' if num_retried == 0 or num_retried == 1
             else ' retries.'), actual_timeout, exception)

    def _handle_retry(self, re, request, throttle_retried):
        throttle_retried += 1
        msg = ('Retry for request ' + request.__class__.__name__ + ', num ' +
               'retries: ' + str(throttle_retried) + ', exception: ' + str(re))
        self._logutils.log_debug(msg)
        handler = self._retry_handler
        if not handler.do_retry(request, throttle_retried, re):
            self._logutils.log_debug(
                'Operation not retry-able or too many retries.')
            raise re
        handler.delay(throttle_retried, re)
        return throttle_retried

    def _log_retried(self, num_retried, exception):
        msg = ('Client, doing retry: ' + str(num_retried) +
               ('' if exception is None else ', exception: ' + str(exception)))
        if (self._request is not None and
                isinstance(exception, SecurityInfoNotReadyException)):
            self._logutils.log_debug(msg)
        else:
            self._logutils.log_info(msg)

    @synchronized
    def _next_request_id(self):
        """
        Get the next client-scoped request id. It needs to be combined with the
        client id to obtain a globally unique scope.
        """
        self._max_request_id += 1
        return self._max_request_id

    def _process_response(self, request, content, status):
        """
        Convert the url_response into a suitable return value.

        :param request: the request executed by the server.
        :type request: Request
        :param content: the content of the response from the server.
        :type content: bytes for python 3 and str for python 2
        :param status: the status code of the response from the server.
        :type status: int
        :returns: the programmatic response object.
        :rtype: Result
        """
        if status == codes.ok:
            bis = ByteInputStream(bytearray(content))
            return self._process_ok_response(bis, request)
        self._process_not_ok_response(content, status)
        raise IllegalStateException('Unexpected http response status: ' +
                                    str(status))

    def _process_ok_response(self, bis, request):
        """
        Process an OK response.

        :param bis: the byte input stream created from the content of response
            get from the server.
        :type bis: ByteInputStream
        :param request: the request executed by the server.
        :type request: Request
        :returns: the result of processing the successful request.
        :rtype: Result
        """
        code = bis.read_byte()
        if code == 0:
            res = request.create_serializer().deserialize(
                request, bis, BinaryProtocol.SERIAL_VERSION)
            if request.is_query_request():
                if not request.is_simple_query():
                    request.get_driver().set_client(self._client)
            return res

        """
        Operation failed. Handle the failure and throw an appropriate
        exception.
        """
        err = BinaryProtocol.read_string(bis)
        raise BinaryProtocol.map_exception(code, err)

    @staticmethod
    def _process_not_ok_response(content, status):
        """
        Process not OK response. The method typically throws an appropriate
        exception. A normal return indicates that the method declined to handle
        the response and it's the caller's responsibility to take appropriate
        action.

        :param content: content of the response from the server.
        :type content: bytes for python 3 and str for python 2
        :param status: the status code of the response from the server.
        :type status: int
        """
        if status == codes.bad:
            length = len(content)
            err_msg = (content if length > 0 else str(status))
            raise NoSQLException('Error response: ' + err_msg)
        raise NoSQLException('Error response = ' + str(status))

    @staticmethod
    def _timeout_request(start_time, request_timeout, sec_timeout_ms,
                         exception):
        """
        Determine if the request should be timed out. If the last exception if
        the SecurityInfoNotReadyException, use its specific timeout to
        determine. Otherwise, check if the request exceed the timeout given.

        :param start_time: when the request starts.
        :type start_time: int
        :param request_timeout: the default timeout of this request.
        :type request_timeout: int
        :param sec_timeout_ms: the timeout of waiting security information to be
            available in milliseconds.
        :type sec_timeout_ms: int
        :param exception: the last exception.
        :type exception: RuntimeError
        :returns: True if the request need to be timed out.
        :rtype: bool
        """
        if isinstance(exception, SecurityInfoNotReadyException):
            return int(round(time() * 1000)) - start_time >= sec_timeout_ms
        return int(round(time() * 1000)) - start_time >= request_timeout



#
# Copyright (C) 2018, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl
#
# Please see LICENSE.txt file included in the top-level directory of the
# appropriate download for a copy of the license and additional information.
#

from logging import DEBUG
from platform import python_version
from requests import Session, adapters
from sys import version_info

from .common import ByteOutputStream, CheckValue, HttpConstants, LogUtils
from .config import DefaultRetryHandler
from .exception import IllegalArgumentException
from .http import RequestUtils
from .operations import QueryResult
from .query import QueryDriver
from .serde import BinaryProtocol
from .version import __version__


class Client(object):
    TRACE_LEVEL = 0

    # The HTTP driver client.
    def __init__(self, config, logger):
        self._logutils = LogUtils(logger)
        self._config = config
        self._url = config.get_service_url()
        self._request_uri = self._url.geturl() + HttpConstants.NOSQL_DATA_PATH
        self._pool_connections = config.get_pool_connections()
        self._pool_maxsize = config.get_pool_maxsize()
        self._max_request_id = 1
        self._proxy_host = config.get_proxy_host()
        self._proxy_port = config.get_proxy_port()
        self._proxy_username = config.get_proxy_username()
        self._proxy_password = config.get_proxy_password()
        self._retry_handler = config.get_retry_handler()
        if self._retry_handler is None:
            self._retry_handler = DefaultRetryHandler()
        self._sec_info_timeout = config.get_sec_info_timeout()
        self._shut_down = False
        self._user_agent = self._make_user_agent()
        self._auth_provider = config.get_authorization_provider()
        if self._auth_provider is None:
            raise IllegalArgumentException(
                'Must configure AuthorizationProvider.')
        self._sess = Session()
        adapter = adapters.HTTPAdapter(pool_connections=self._pool_connections,
                                       pool_maxsize=self._pool_maxsize,
                                       max_retries=5, pool_block=True)
        self._sess.mount(self._url.scheme + '://', adapter)
        if self._proxy_host is not None:
            self._check_and_set_proxy(self._sess)

    def execute(self, request):
        """
        Execute the KV request and return the response. This is the top-level
        method for request execution.

        This method handles exceptions to distinguish between what can be
        retried and what cannot, making sure that root cause exceptions are
        kept. Examples:

            can't connect (host, port, etc)\n
            throttling exceptions\n
            general networking issues, IOError\n

        RequestTimeoutException needs a cause, or at least needs to include the
        message from the causing exception.

        :param request: the request to be executed by the server.
        :type request: Request
        :returns: the result of the request.
        :rtype: Result
        :raises IllegalArgumentException: raises the exception if request is
            None.
        """
        CheckValue.check_not_none(request, 'request')
        request.set_defaults(self._config)
        request.validate()
        if request.is_query_request():
            """
            The following 'if' may be True for advanced queries only. For such
            queries, the 'if' will be True (i.e., the QueryRequest will be bound
            with a QueryDriver) if and only if this is not the 1st execute()
            call for this query. In this case we just return a new, empty
            QueryResult. Actual computation of a result batch will take place
            when the app calls get_results() on the QueryResult.
            """
            if request.has_driver():
                self._trace('QueryRequest has QueryDriver', 2)
                return QueryResult(request, False)
            """
            If it is an advanced query and we are here, then this must be the
            1st execute() call for the query. If the query has been prepared
            before, we create a QueryDriver and bind it with the QueryRequest.
            Then, we create and return an empty QueryResult. Actual computation
            of a result batch will take place when the app calls get_results()
            on the QueryResult.
            """
            if request.is_prepared() and not request.is_simple_query():
                self._trace(
                    'QueryRequest has no QueryDriver, but is prepared', 2)
                driver = QueryDriver(request)
                driver.set_client(self)
                driver.set_topology_info(request.topology_info())
                return QueryResult(request, False)
            """
            If we are here, then this is either (a) a simple query or (b) an
            advanced query that has not been prepared already, which also
            implies that this is the 1st execute() call on this query. For a
            non-prepared advanced query, the effect of this 1st execute() call
            is to send the query to the proxy for compilation, get back the
            prepared query, but no query results, create a QueryDriver, and bind
            it with the QueryRequest (see QueryRequestSerializer.deserialize()),
            and return an empty QueryResult.
            """
            self._trace(
                'QueryRequest has no QueryDriver and is not prepared', 2)
        timeout_ms = request.get_timeout()
        content = self._write_content(request)
        BinaryProtocol.check_request_size_limit(request, len(content))
        headers = {'Host': self._url.hostname,
                   'Content-Type': 'application/octet-stream',
                   'Connection': 'keep-alive',
                   'Accept': 'application/octet-stream',
                   'Content-Length': str(len(content)),
                   'User-Agent': self._user_agent}
        if request.get_compartment() is None:
            request.set_compartment_internal(
                self._config.get_default_compartment())
        if self._logutils.is_enabled_for(DEBUG):
            self._logutils.log_debug('Request: ' + request.__class__.__name__)
        request_utils = RequestUtils(
            self._sess, self._logutils, request, self._retry_handler, self)
        return request_utils.do_post_request(
            self._request_uri, headers, content, timeout_ms,
            self._sec_info_timeout)

    def get_auth_provider(self):
        return self._auth_provider

    def shut_down(self):
        # Shutdown the client.
        self._logutils.log_info('Shutting down driver http client')
        if self._shut_down:
            return
        self._shut_down = True
        if self._auth_provider is not None:
            self._auth_provider.close()
        if self._sess is not None:
            self._sess.close()

    def _check_and_set_proxy(self, sess):
        if (self._proxy_host is not None and self._proxy_port == 0 or
                self._proxy_host is None and self._proxy_port != 0):
            raise IllegalArgumentException(
                'To configure an HTTP proxy, both host and port are required.')
        if (self._proxy_username is not None and self._proxy_password is None or
                self._proxy_username is None and
                self._proxy_password is not None):
            raise IllegalArgumentException(
                'To configure HTTP proxy authentication, both user name and ' +
                'password are required')
        if self._proxy_host is not None:
            if self._proxy_username is None:
                proxy_url = (
                    'http://' + self._proxy_host + ':' + str(self._proxy_port))
            else:
                assert self._proxy_password is not None
                proxy_url = (
                    'http://' + self._proxy_username + ':' +
                    self._proxy_password + '@' + self._proxy_host + ':' +
                    str(self._proxy_port))
            sess.proxies = {'http': proxy_url, 'https': proxy_url}

    @staticmethod
    def _make_user_agent():
        if version_info.major >= 3:
            pyversion = python_version()
        else:
            pyversion = '%s.%s.%s' % (version_info.major, version_info.minor,
                                      version_info.micro)
        return '%s/%s (Python %s)' % ('NoSQL-PythonSDK', __version__, pyversion)

    @staticmethod
    def _trace(msg, level):
        if level <= Client.TRACE_LEVEL:
            print('DRIVER: ' + msg)

    @staticmethod
    def _write_content(request):
        """
        Serializes the request payload, sent as http content.

        :param request: the request to be executed by the server.
        :type request: Request
        :returns: the bytearray that contains the content.
        :rtype: bytearray
        """
        content = bytearray()
        bos = ByteOutputStream(content)
        BinaryProtocol.write_serial_version(bos)
        request.create_serializer().serialize(
            request, bos, BinaryProtocol.SERIAL_VERSION)
        return content



#
# Copyright (C) 2018, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl
#
# Please see LICENSE.txt file included in the top-level directory of the
# appropriate download for a copy of the license and additional information.
#

__version__ = '5.2.0'
